package ru.dbotthepony.mc.otm.container.util

import net.minecraft.world.Container
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.get
import ru.dbotthepony.mc.otm.container.set
import ru.dbotthepony.mc.otm.core.GetterSetter
import ru.dbotthepony.mc.otm.core.collect.filter
import ru.dbotthepony.mc.otm.core.collect.map
import ru.dbotthepony.mc.otm.core.isNotEmpty

interface IContainerSlot : GetterSetter<ItemStack> {
	val slot: Int
	val container: Container

	operator fun component1() = slot
	operator fun component2() = item

	val isForbiddenForAutomation: Boolean get() {
		val container = container

		if (container is MatteryContainer) {
			return container.isSlotForbiddenForAutomation(slot)
		} else {
			return false
		}
	}

	val filter: Item? get() {
		val container = container

		if (container is MatteryContainer) {
			return container.getSlotFilter(slot)
		} else {
			return null
		}
	}

	val hasFilter: Boolean
		get() = filter != null

	fun remove() {
		container[slot] = ItemStack.EMPTY
	}

	override fun get(): ItemStack {
		return container[slot]
	}

	override fun accept(t: ItemStack) {
		container[slot] = t
	}

	fun setChanged() {
		val container = container

		if (container is MatteryContainer) {
			container.setChanged(slot)
		} else {
			container.setChanged()
		}
	}

	var item: ItemStack
		get() = container[slot]
		set(value) { container[slot] = value }

	val isEmpty: Boolean
		get() = container[slot].isEmpty

	val isNotEmpty: Boolean
		get() = container[slot].isNotEmpty
}

class ContainerSlot(override val slot: Int, override val container: Container) : IContainerSlot {
	init {
		require(slot in 0 until container.containerSize) { "Slot out of bounds: $slot (container: $container with size ${container.containerSize})" }
	}
}

interface IIterableContainer : Iterable<ItemStack> {
	/**
	 * Iterates over non-empty itemstacks of this container
	 */
	override fun iterator(): Iterator<ItemStack> {
		return iterator(true)
	}

	fun iterator(nonEmpty: Boolean): Iterator<ItemStack>

	/**
	 * Iterates non-empty slots of this container
	 */
	fun slotIterator(): Iterator<IContainerSlot> {
		return slotIterator(true)
	}

	fun slotIterator(nonEmpty: Boolean): Iterator<IContainerSlot>
}

operator fun Container.iterator() = iterator(true)

fun Container.iterator(nonEmpty: Boolean): Iterator<ItemStack> {
	if (this is IIterableContainer) {
		return iterator(nonEmpty)
	} else if (nonEmpty) {
		return (0 until containerSize).iterator().map { this[it] }.filter { it.isNotEmpty }
	} else {
		return (0 until containerSize).iterator().map { this[it] }
	}
}

fun Container.slotIterator(nonEmpty: Boolean = true): Iterator<IContainerSlot> {
	if (this is IIterableContainer) {
		return slotIterator(nonEmpty)
	} else if (nonEmpty) {
		return (0 until containerSize).iterator().filter { this[it].isNotEmpty }.map { ContainerSlot(it, this) }
	} else {
		return (0 until containerSize).iterator().map { ContainerSlot(it, this) }
	}
}
